import os

from typing import Dict, List, Any
from metaflow import Task, namespace, Step
from metaflow.plugins.env_escape import generate_trampolines


class DebugStubGenerator(object):
    def __init__(self, task_pathspec):
        """
        Initializes the DebugStubGenerator.

        Parameters
        ----------
        task_pathspec : str
            The pathspec of the task.
        """
        self.task_pathspec = task_pathspec
        self.task = Task(task_pathspec, _namespace_check=False)
        self.step_name = self.task.parent.id
        self.run_id = self.task.parent.parent.id
        self.flow_name = self.task.parent.parent.parent.id
        self.is_new_conda_step = self.is_new_conda_step()

        parameters_task = Step(
            "%s/_parameters" % self.task.parent.parent.pathspec, _namespace_check=False
        ).task

        self.workflow_dag = parameters_task["_graph_info"].data
        self.file_name = self.workflow_dag["file"]
        self._dag_structure = self.get_dag_structure(self.workflow_dag["steps"])
        self.step_type = self.get_step_type(self.step_name)
        self._previous_nodes = self.find_previous_nodes(
            self.step_name, self._dag_structure
        )
        self.task_line_num = self._dag_structure[self.step_name]["line"]
        self.node = self._dag_structure[self.step_name]
        self.previous_steps = self.get_previous_steps(self._previous_nodes)
        self.previous_tasks = self.get_previous_tasks(self.previous_steps)
        self.task_namespace = self.get_task_namespace()

    @staticmethod
    def get_dag_structure(dag: Dict[str, Any]) -> Dict[str, Any]:
        """
        Returns the simplified DAG structure of the workflow.

        Parameters:
        ----------
        dag : Dict[str, Any]
            The DAG structure of the workflow.
        Returns:
        ----------
        Dict[str, Any]
            Simplified DAG structure of the workflow.
        """
        dag_structure = {}
        for node, attributes in dag.items():
            dag_structure[node] = {
                "next": attributes["next"],
                "type": attributes["type"],
                "line": attributes["line"],
            }
        return dag_structure

    def is_new_conda_step(self) -> bool:
        """
        Returns True if the step is a new conda step, False otherwise.

        Returns:
        ----------
        bool
            True if the step is a new conda step, False otherwise.
        """
        return "conda_env_id" in self.task.metadata_dict

    def get_step_type(self, step_name: str) -> str:
        """
        Returns the type of the step.

        Parameters
        ----------
        step_name : str
            The name of the step.

        Returns
        -------
        str
            The type of the step.
        """
        return self._dag_structure[step_name]["type"]

    @staticmethod
    def find_previous_nodes(node: str, dag_structure: Dict[str, Any]) -> List[str]:
        """
        Returns the previous nodes for a given node.

        Parameters
        ----------
        node : str
            The name of the node to find the previous nodes for.
        dag_structure : Dict[str, Any]
            The DAG structure of the workflow.

        Returns
        -------
        List[str]
            The list of previous nodes.
        """
        if node == "start":
            return []
        previous_nodes = []
        for node_name, attributes in dag_structure.items():
            if node in attributes["next"]:
                previous_nodes.append(node_name)
        return previous_nodes

    @staticmethod
    def get_join_type(previous_steps: List[Step]) -> str:
        """
        Returns the type of join for the step.

        Parameters
        ----------
        previous_steps : List[Step]
            The list of previous steps.

        Returns
        -------
        str
            The type of join for the step.
        """
        return "foreach" if len(previous_steps) == 1 else "static"

    def get_previous_steps(self, previous_nodes: List[str]) -> List[Step]:
        """
        Returns the previous steps for a given step.

        Parameters
        ----------
        previous_nodes : List[str]
            The list of previous nodes.

        Returns
        -------
        List[Step]
            The list of previous steps.
        """
        return [
            Step(
                "{}/{}/{}".format(self.flow_name, self.run_id, n),
                _namespace_check=False,
            )
            for n in previous_nodes
        ]

    def get_previous_tasks(self, previous_steps: List[Step]) -> List[Task]:
        """
        Returns the previous tasks for a given step.

        Parameters
        ----------
        previous_steps : List[Step]
            The list of previous steps.

        Returns
        -------
        List[Task]
            The list of previous tasks.
        """
        step_type = self.get_step_type(self.step_name)
        if step_type == "join" and self.get_join_type(self.previous_steps) == "foreach":
            return sorted(
                [task for task in previous_steps[0].tasks()], key=lambda x: x.index
            )
        # TODO: This method is incomplete and incorrect in the general case. We actually
        # need more information to return the exact subset of tasks that are runtime
        # parents in some situations with nested foreaches. For now, best effort but
        # we need to revisit/fix these. The primary limitation right now is that the
        # `foreach-stack` field we are referring to is capped at some number of characters
        # which can fail to work for deeply nested foreaches or even shallow ones with
        # long values.
        foreach_list = self.task.metadata_dict.get("foreach-stack", [])
        if not foreach_list:
            # We are not part of a foreach so return all the previous tasks. This will
            # be a list of 1 for most everything except for a join.
            return [step.task for step in previous_steps]

        # We are part of a foreach, we want to list tasks that either have the same
        # foreach_list or match everything but the last element
        def _filter(t):
            t_foreach_list = t.metadata_dict.get("foreach-stack", [])
            return (
                len(t_foreach_list) == len(foreach_list)
                and t_foreach_list == foreach_list
            ) or t_foreach_list == foreach_list[:-1]

        to_return = []
        for step in previous_steps:
            to_return.extend([task for task in step if _filter(task)])
        return to_return

    def get_task_namespace(self) -> str:
        """
        Returns the namespace of the task.

        Returns
        -------
        str
            The namespace of the task.
        """
        task_namespace = None
        for val in self.task.tags:
            if val.startswith("production:") or val.startswith("user:"):
                task_namespace = val
                break
        return task_namespace



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import json
import os
import sys
import tempfile

from metaflow._vendor.packaging.utils import canonicalize_version

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,  # noqa
    Set,  # noqa
    Union,  # noqa
    cast,
)

from metaflow.datastore.flow_datastore import FlowDataStore
from metaflow.datastore.task_datastore import TaskDataStore
from metaflow.decorators import StepDecorator
from metaflow.flowspec import FlowSpec
from metaflow.graph import FlowGraph
from metaflow.metadata_provider import MetaDatum, MetadataProvider
from metaflow.metaflow_config import CONDA_REMOTE_COMMANDS
from metaflow.metaflow_environment import (
    InvalidEnvironmentException,
    MetaflowEnvironment,
)

from metaflow.packaging_sys import ContentType
from metaflow.plugins.env_escape import generate_trampolines
from metaflow.unbounded_foreach import UBF_TASK

from .conda_environment import CondaEnvironment
from .env_descr import EnvID
from .utils import arch_id

from .conda_common_decorator import (
    CondaRequirementDecoratorMixin,
    NamedEnvRequirementDecoratorMixin,
    PypiRequirementDecoratorMixin,
    SysPackagesRequirementDecoratorMixin,
    StepRequirementMixin,
)
from .conda import Conda


class PackageRequirementStepDecorator(StepDecorator):
    name = "step_package_req"

    allow_multiple = True

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *@%s* decorator requires " "--environment=conda" % self.name
            )


class CondaRequirementStepDecorator(
    CondaRequirementDecoratorMixin, PackageRequirementStepDecorator
):
    """
    Specifies the Conda packages for the step.

    Information in this decorator will augment any
    attributes set in the `@conda_base`, `@pypi_base` or `@named_env_base`
    flow-level decorator. Hence you can use the flow decorators to set common libraries
    required by all steps and use `@conda`, `@pypi` to specify step-specific additions
    or replacements.
    Information specified in this decorator will augment the information in the base
    decorator and, in case of a conflict (for example the same library specified in
    both the base decorator and the step decorator), the step decorator's information
    will prevail.

    Parameters
    ----------
    libraries : Dict[str, str], default {}
        Libraries to use for this step. The key is the name of the package
        and the value is the version to use. Note that versions can
        be specified either as a specific version or as a comma separated string
        of constraints like "<2.0,>=1.5".
    channels : List[str], default []
        Additional channels to search
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current version
        will be used.
    disabled : bool, default False
        If set to True, uses the external environment.
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(pathspec=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the pathspec, you can use `@{}` values and
        environment variables will be used to substitute.
    pip_packages : Dict[str, str], default {}
        DEPRECATED -- use `@pypi(packages=)` instead.
        Same as libraries but for pip packages.
    pip_sources : List[str], default []
        DEPRECATED -- use `@pypi(extra_indices=)` instead.
        Same as channels but for pip sources.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(fetch_at_exec=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    """

    name = "conda"

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        deprecated_keys = set(
            ("pip_packages", "pip_sources", "fetch_at_exec", "name", "pathspec")
        ).intersection((k for k, v in self.attributes.items() if v))
        if deprecated_keys:
            logger(
                "*DEPRECATED*: Using '%s' in '@%s' is deprecated. Please use '@pypi' or "
                "'@named_env' instead. " % (", ".join(deprecated_keys), self.name)
            )
        return super().step_init(
            flow, graph, step_name, decorators, environment, flow_datastore, logger
        )


class PypiRequirementStepDecorator(
    PypiRequirementDecoratorMixin, PackageRequirementStepDecorator
):
    """
    Specifies the Pypi packages for the step.

    Information in this decorator will augment any
    attributes set in the `@conda_base`, `@pypi_base` or `@named_env_base`
    flow-level decorator. Hence you can use the flow decorators to set common libraries
    required by all steps and use `@conda`, `@pypi` to specify step-specific additions
    or replacements.
    Information specified in this decorator will augment the information in the base
    decorator and, in case of a conflict (for example the same library specified in
    both the base decorator and the step decorator), the step decorator's information
    will prevail.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        Packages to use for this step. The key is the name of the package
        and the value is the version to use (default `{}`).
    conda_only: Dict[str, str], default {}
        Packages that are only available in conda. These packages should not depend on
        any python package (example ffmpeg). Note that this is different than using
        the `@conda_base` decorator as using `conda_only` will still build a Pypi
        environment instead of a mixed one.
    extra_indices : List[str], default []
        Additional sources to search for
    extras: List[str], default []
        Extra arguments to pass to the resolver. For example "pre" to include
        pre-release packages.
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current python
        version will be used.
    disabled : bool, default False
        If set to True, uses the external environment.
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(name=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    """

    name = "pypi"

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        deprecated_keys = set(
            ("sources", "fetch_at_exec", "name", "pathspec")
        ).intersection((k for k, v in self.attributes.items() if v))

        if deprecated_keys:
            logger(
                "*DEPRECATED*: Using '%s' in '@%s' is deprecated. Please use "
                "'@named_env' instead. " % (", ".join(deprecated_keys), self.name)
            )
        return super().step_init(
            flow, graph, step_name, decorators, environment, flow_datastore, logger
        )


class NamedEnvRequirementStepDecorator(
    NamedEnvRequirementDecoratorMixin, PackageRequirementStepDecorator
):
    """
    Specifies a named environment to extract the environment from

    Information in this decorator will augment any
    attributes set in the `@conda_base`, `@pypi_base` or `@named_env_base`
    flow-level decorator. Hence you can use the flow decorators to set common libraries
    required by all steps and use `@conda`, `@pypi` to specify step-specific additions
    or replacements.
    Information specified in this decorator will augment the information in the base
    decorator and, in case of a conflict (for example the same library specified in
    both the base decorator and the step decorator), the step decorator's information
    will prevail.

    Parameters
    ----------
    name : str, optional, default None
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "named_env"


class SysPackagesRequirementStepDecorator(
    SysPackagesRequirementDecoratorMixin, PackageRequirementStepDecorator
):
    """
    Specifies system virtual packages for this step.

    This is an advanced usage decorator allowing you to override the __glibc and
    __cuda virtual packages that are used when resolving your environment.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        System virtual packages to use for this flow. Supported keys are "__cuda" and
        "__glibc".
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "sys_packages"


# Here for legacy reason -- use @pypi instead
class PipRequirementStepDecorator(PypiRequirementStepDecorator):
    """
    Specifies the Pypi packages for the step.

    DEPRECATED: please use `@pypi` instead.

    Parameters
    ----------
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    packages : Dict[str, str], default {}
        Packages to use for this step. The key is the name of the package
        and the value is the version to use (default `{}`).
    extra_indices : List[str], default []
        Additional sources to search for
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current python
        version will be used.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(name=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "pip"

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        logger("*DEPRECATED*: Use '@pypi' instead of '@%s'." % self.name)
        return super().step_init(
            flow, graph, step_name, decorators, environment, flow_datastore, logger
        )


class PylockTomlInternalDecorator(StepRequirementMixin, StepDecorator):
    defaults = {
        "path": None,
        "disabled": None,
        # Context for user_deps_for_hash attribute:
        #
        # * For reusing hydrated environments, we were using
        #   env_descr.py::get_req_id() to calculate the id (actually hash) for
        #   each step's environment
        # * In this calculation, three dictionaries: deps (user_deps), sources,
        #   extras, are considered.
        #
        # And why we have this attribute here:
        #
        # * We chose not to add the hash of file_paths to the environment hash
        #   (req_id).
        # * We added a user_deps_for_hash dict parameter for the
        #   pylock_toml_internal decorator.
        # * We deliberately named this parameter not to be user_dep to hint this
        #   passed in dep is only used for hash calculation, not for actual
        #   dependencies resolution.
        # * Only the content in passed pylock_toml file is respected to
        #   determine the actual packages to be resolved.
        # * Caveat: if there are inconsistencies between the user_dep and
        #   pylock_toml, such as a user package is in the resulting pylock.toml
        #   but not specified in user_dep, we could end up reusing the wrong
        #   hydrated environment. And this kind of bug is very hard to locate.
        #
        # Mitigations: 1) we will carefully use this internal decorator with this
        # knowledge, only for the seed users. 2) We will have robust test on the
        # upperstream @uv decorator, which will assign the user_deps_for_hash.
        "user_deps_for_hash": {},
        # The purpose is similar to user_deps_for_hash.
        "user_sources_for_hash": [],
        # When a pylock_internal decorator is generated, the originating
        # ResolvedUvEnvFlowDecorator calculates the hash of uv.lock, assigns it
        # to the decorator's attributes["full_id_unique_keys"]["uv_content_hash"],
        # and eventually this uv content hash value will be passed to
        # PylockTomlInternalDecorator.resolve()'s parameters.
        #
        # This is used to detect a uv.lock change with unchanged user
        # dependencies in pyproject.toml, caused by design by uv's non-determinism
        # in resolution when ">=version" specifiers are involved.
        "full_id_unique_keys": {},
    }

    name = "pylock_toml_internal"

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "Decorators specifying flow or step dependency environments require --environment=conda",
            )

        return super().step_init(
            flow, graph, step_name, decorators, environment, flow_datastore, logger
        )

    @property
    def file_paths(self) -> Dict[str, List[str]]:
        return {
            # TODO
            # more robust handling of a missing pylock_toml key in the requirements path.
            "pylock_toml": [self.attributes["path"]],
        }

    @property
    # See the notes at the definition of user_deps_for_hash for more contexts of this.
    def packages(self) -> Dict[str, Dict[str, str]]:
        return {
            "pypi": {
                k: canonicalize_version(v)
                for k, v in cast(
                    Dict[str, str], self.attributes["user_deps_for_hash"]
                ).items()
            },
        }

    @property
    def sources(self) -> Dict[str, List[str]]:
        return {
            "pypi": self.attributes["user_sources_for_hash"],
        }

    @property
    def python(self) -> Optional[str]:
        return self.attributes["user_deps_for_hash"].get("python")

    @property
    def full_id_unique_keys(self) -> Dict[str, str]:
        return cast(Dict[str, str], self.attributes["full_id_unique_keys"])


class CondaEnvInternalDecorator(StepDecorator):
    name = "conda_env_internal"
    TYPE = "conda"

    conda = None  # type: Optional[Conda]
    _local_root = None  # type: Optional[str]

    _metaflow_home = None  # type: Optional[str]
    _addl_env_vars = None  # type: Optional[Dict[str, str]]

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        self._echo = logger
        self._env = cast(CondaEnvironment, environment)
        self._flow = flow
        self._step_name = step_name
        self._flow_datastore = flow_datastore

        # Environment variables used in resolving at fetch time pathspec/name
        self._env_for_fetch = {}  # type: Dict[str, Union[str, Callable[[], str]]]

        self._env_id = None  # type: Optional[EnvID]

        self._is_remote = any([d.name in CONDA_REMOTE_COMMANDS for d in decorators])

        os.environ["PYTHONNOUSERSITE"] = "1"

    def runtime_init(self, flow: FlowSpec, graph: FlowGraph, package: Any, run_id: str):

        if self.__class__._metaflow_home is None:
            # Do this ONCE per flow (thus class variable)
            # Create a symlink to installed version of metaflow to execute user code against
            self.__class__._metaflow_home = tempfile.mkdtemp(dir="/tmp")
            package.extract_into(
                self.__class__._metaflow_home,
                ContentType.CODE_CONTENT.value
                | ContentType.MODULE_CONTENT.value
                | ContentType.OTHER_CONTENT.value,
            )

            self.__class__._addl_env_vars = package.get_post_extract_env_vars(
                package.package_metadata, self.__class__._metaflow_home
            )

            # Also install any environment escape overrides directly here to enable
            # the escape to work even in non metaflow-created subprocesses
            generate_trampolines(self.__class__._metaflow_home)

        # If we need to fetch the environment on exec, save the information we need
        # so that we can resolve it using information such as run id, step name, task
        # id and parameter values
        if self._is_enabled() and self._is_fetch_at_exec():
            self._env_for_fetch["METAFLOW_RUN_ID"] = run_id
            self._env_for_fetch["METAFLOW_RUN_ID_BASE"] = run_id
            self._env_for_fetch["METAFLOW_STEP_NAME"] = self._step_name

    def runtime_task_created(
        self,
        task_datastore: TaskDataStore,
        task_id: str,
        split_index: int,
        input_paths: List[str],
        is_cloned: bool,
        ubf_context: str,
    ):
        if self._is_enabled(ubf_context):
            if self._is_fetch_at_exec():
                # We need to ensure we can properly find the environment we are
                # going to run in
                run_id, step_name, task_id = input_paths[0].split("/")
                parent_ds = self._flow_datastore.get_task_datastore(
                    run_id, step_name, task_id
                )
                for var, param in self._flow._get_parameters():
                    if param.IS_CONFIG_PARAMETER:
                        continue
                    self._env_for_fetch[
                        "METAFLOW_INIT_%s" % var.upper().replace("-", "_")
                    ] = lambda _param=getattr(  # type: ignore[misc]
                        self._flow, var
                    ), _var=var, _ds=parent_ds: str(
                        _param.load_parameter(_ds[_var])
                    )
                self._env_for_fetch["METAFLOW_TASK_ID"] = task_id

                self._env_id = self._env.resolve_fetch_at_exec_env(
                    self._step_name, self._env_for_fetch
                )
                if self._env_id is None:
                    raise InvalidEnvironmentException(
                        "Cannot find the environment ID for a fetch-at-exec step"
                    )
            else:
                t = self._env.get_env_id_noconda(self._step_name)
                if isinstance(t, EnvID):
                    self._env_id = t
                else:
                    raise InvalidEnvironmentException(
                        "Unexpected ID for the Conda environment for step '%s': '%s'"
                        % (self._step_name, str(t))
                    )

    def runtime_step_cli(
        self,
        cli_args: Any,  # Importing CLIArgs causes an issue so ignore for now
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
    ):
        # Check if there is a conda environment to use. We check for UBF_TASK (instead of
        # ubf_context) because, even if we don't create the environment for a control task,
        # we want to pass down the environment variable for the mapper tasks.
        if self._is_enabled(UBF_TASK):
            cli_args.env["_METAFLOW_CONDA_ENV"] = json.dumps(self._env_id)
            if self._is_remote or not self._is_enabled(ubf_context):
                # If we are executing remotely, we don't need to actually create the
                # environment at all. We also don't create it if we are local and
                # this is not enabled for the current UBF context.
                return
        else:
            # In this case, there is no need for a conda environment. Since we can
            # be running with a runner *inside* another Metaflow step that *does* have
            # a conda environment, we make sure to clear _METAFLOW_CONDA_ENV so that
            # the outside environment doesn't "leak" (it could have been set for the external
            # process in which case it would get passed down).
            cli_args.env["_METAFLOW_CONDA_ENV"] = ""
            return

        # Create the environment only if executing locally (ie: not self._is_remote)
        conda = cast(Conda, self._env.conda)
        assert self._env_id
        entrypoint = None  # type: Optional[str]
        # Create the environment we are going to use
        existing_env_info = conda.created_environment(self._env_id)
        if existing_env_info:
            self._echo(
                "Using existing Conda environment %s (%s)"
                % (self._env_id.req_id, self._env_id.full_id)
            )
            entrypoint = os.path.join(existing_env_info[1], "bin", "python")
        else:
            # Otherwise, we read the conda file and create the environment locally
            self._echo(
                "Creating Conda environment %s (%s)..."
                % (self._env_id.req_id, self._env_id.full_id)
            )
            resolved_env = conda.environment(self._env_id)
            if resolved_env:
                entrypoint = os.path.join(
                    conda.create_for_step(self._step_name, resolved_env),
                    "bin",
                    "python",
                )
            else:
                raise InvalidEnvironmentException("Cannot create environment")

        # Actually set it up.
        python_path = self.__class__._metaflow_home
        addl_env_vars = {}
        if self.__class__._addl_env_vars:
            for key, value in self.__class__._addl_env_vars.items():
                if key.endswith(":"):
                    addl_env_vars[key[:-1]] = value
                elif key == "PYTHONPATH":
                    path_components = [value]
                    if python_path is not None:
                        path_components.append(python_path)
                    addl_env_vars[key] = os.pathsep.join(path_components)
                else:
                    addl_env_vars[key] = value
        cli_args.env.update(addl_env_vars)

        if entrypoint is None:
            # This should never happen -- it means the environment was not
            # created somehow
            raise InvalidEnvironmentException("No executable found for environment")

        if arch_id().startswith("linux"):
            # No need for MacOS -- with SIP, DYLD_LIBRARY_PATH is ignored anyways
            old_ld_path = os.environ.get("LD_LIBRARY_PATH")
            if old_ld_path is not None:
                cli_args.env["MF_ORIG_LD_LIBRARY_PATH"] = old_ld_path
                cli_args.env["LD_LIBRARY_PATH"] = ":".join(
                    [
                        os.path.join(
                            os.path.dirname(os.path.dirname(entrypoint)), "lib"
                        ),
                        old_ld_path,
                    ]
                )
        cli_args.entrypoint[0] = entrypoint

        if os.environ.get("METAFLOW_COVERAGE_S3_PATH"):
            from metaflow_extensions.netflix_ext.plugins.coverage.setup_coverage import (
                setup_coverage,
            )

            setup_coverage(entrypoint)

    def task_pre_step(
        self,
        step_name: str,
        task_datastore: TaskDataStore,
        metadata: MetadataProvider,
        run_id: str,
        task_id: str,
        flow: FlowSpec,
        graph: FlowGraph,
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
        inputs: List[str],
    ):
        if self._is_enabled(ubf_context):
            # Add the Python interpreter's parent to the path. This is to
            # ensure that any non-pythonic dependencies introduced by the conda
            # environment are visible to the user code.
            env_path = os.path.dirname(os.path.realpath(sys.executable))
            if os.environ.get("PATH") is not None:
                env_path = os.pathsep.join([env_path, os.environ["PATH"]])
            os.environ["PATH"] = env_path

            metadata.register_metadata(
                run_id,
                step_name,
                task_id,
                [
                    MetaDatum(
                        field="conda_env_id",
                        value=os.environ["_METAFLOW_CONDA_ENV"],
                        type="conda_env_id",
                        tags=["attempt_id:{0}".format(retry_count)],
                    )
                ],
            )

    def runtime_finished(self, exception: Exception):
        pass
        # if self.__class__._metaflow_home is not None:
        #     shutil.rmtree(self.__class__._metaflow_home)
        #     self.__class__._metaflow_home = None

    def _is_enabled(self, ubf_context: str = UBF_TASK) -> bool:
        return CondaEnvironment.enabled_for_step(self._step_name, ubf_context)

    def _is_fetch_at_exec(self) -> bool:
        return CondaEnvironment.fetch_at_exec_for_step(self._step_name)



import time
from functools import wraps


def retry_exp_backoff(
    exceptions_to_check,
    user_msg,
    user_facing_exception=None,
    logger=print,
    tries=6,
    delay=1,
    exponential_backoff_factor=2,
    max_delay_per_retry=10,
):
    """
    Exponential backoff based retry, pass in a tuple of Exceptions on which to retry. All time based
    arguments like ('delay' and 'max_delay_per_retry') are in seconds. By default the logger logs
    to stdout (print).

    Always try to pass 'user_facing_exception' in order to give a nice user facing exceptions in
    the end if retrying does not help.

    Args:
        exceptions_to_check (tuple[Exception]): (required) exceptions on which to retry
        user_msg (str): (required) use facing error message that shows during every retry attempt
        user_facing_exception (Exception): (optional) user facing exception in case all retries fail
        logger (Callable): (optional) callable to print the above 'user_msg' and the final exception
        tries (int): (optional) number of retries
        delay (int): (optional) base delay between retries in seconds
        exponential_backoff_factor (int): (optional) exponential backoff factory between retries
        max_delay_per_retry (int): (optional) max delay between single retries in seconds
    """
    if not isinstance(exceptions_to_check, tuple):
        raise ValueError("'exceptions_to_check' should be a 'tuple' of 'Exception'.")

    if not user_msg:
        raise ValueError("'user_msg' must be specified")

    if user_facing_exception and not isinstance(user_facing_exception, Exception):
        raise ValueError(
            "'user_facing_exception' should be a 'Exception' not '%s'"
            % user_facing_exception.__class__.__name__
        )

    def deco_retry(f):
        @wraps(f)
        def retry_func(*args, **kwargs):
            remaining_tries, retry_delay = tries, min(delay, max_delay_per_retry)

            # Attempt to retry with exponential backoff
            while remaining_tries >= 1:
                try:
                    return f(*args, **kwargs)
                except exceptions_to_check as function_exception:

                    if logger:
                        msg = "%s. Retrying in %d seconds..." % (
                            user_msg,
                            retry_delay,
                        )
                        logger(msg)

                    # All retries failed
                    if remaining_tries == 1:
                        logger(str(function_exception))
                        raise user_facing_exception or function_exception

                    time.sleep(retry_delay)

                    # Update params for the next retry
                    remaining_tries -= 1
                    retry_delay *= exponential_backoff_factor
                    retry_delay = min(retry_delay, max_delay_per_retry)

        return retry_func

    return deco_retry



import re
from typing import Any, Dict, List, Optional

from metaflow.metaflow_config import CONDA_SYS_DEPENDENCIES
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow._vendor.packaging.requirements import InvalidRequirement, Requirement


REQ_SPLIT_LINE = re.compile(r"([^~<=>]*)([~<=>]+.*)?")

# Allows things like:
# pkg = <= version
# pkg <= version
# pkg = version
# pkg = ==version or pkg = =version
# In other words, the = is optional but possible
YML_SPLIT_LINE = re.compile(r"(?:=\s)?(<=|>=|~=|==|<|>|=)")


def req_parser(config_value: str) -> Dict[str, Any]:
    extra_args: Dict[str, Any] = {}
    sources: Dict[str, List[str]] = {}
    deps: Dict[str, str] = {}
    np_deps: Dict[str, str] = {}
    sys_deps: Dict[str, str] = {}
    python_version = parse_req_value(
        config_value, extra_args, sources, deps, np_deps, sys_deps
    )
    result: Dict[str, Any] = {}
    if python_version:
        result["python"] = python_version

    if extra_args:
        result["extras"] = extra_args
    if np_deps:
        result["conda_only"] = np_deps
    if sys_deps:
        raise InvalidEnvironmentException(
            "System dependencies are not supported when parsing requirements.txt for "
            "the pypi decorator -- use a named environment instead"
        )

    if "pypi" in sources:
        result["extra_indices"] = sources["pypi"]  # type: ignore[assignment]
        del sources["pypi"]
    if len(sources):
        raise InvalidEnvironmentException(
            "Only PYPI sources are allowed in requirements.txt"
        )

    result["packages"] = deps  # type: ignore[assignment]

    return result


def yml_parser(config_value: str) -> Dict[str, Any]:
    sources: Dict[str, List[str]] = {}
    conda_deps: Dict[str, str] = {}
    pypi_deps: Dict[str, str] = {}
    sys_deps: Dict[str, str] = {}
    python_version = parse_yml_value(
        config_value, {}, sources, conda_deps, pypi_deps, sys_deps
    )
    result = {}
    if sys_deps:
        raise InvalidEnvironmentException(
            "System dependencies are not supported when parsing environment.yml for "
            "the conda decorator -- use a named environment instead"
        )

    if python_version:
        result["python"] = python_version

    if "conda" in sources:
        result["channels"] = sources["conda"]  # type: ignore[assignment]
    if "pypi" in sources:
        result["pip_sources"] = sources["pypi"]  # type: ignore[assignment]

    if len(conda_deps):
        result["libraries"] = conda_deps  # type: ignore[assignment]
    if len(pypi_deps):
        result["pip_packages"] = pypi_deps  # type: ignore[assignment]

    return result


def toml_parser(config_value: str) -> Dict[str, Any]:
    extra_args: Dict[str, List[str]] = {}
    sources: Dict[str, List[str]] = {}
    deps: Dict[str, str] = {}
    np_deps: Dict[str, str] = {}
    sys_deps: Dict[str, str] = {}
    python_version = parse_toml_value(
        config_value, extra_args, sources, deps, np_deps, sys_deps
    )
    result = {}
    if python_version:
        result["python"] = python_version

    if "pypi" in sources:
        result["extra_indices"] = sources["pypi"]  # type: ignore[assignment]
        del sources["pypi"]
    if len(sources):
        raise InvalidEnvironmentException(
            "Only PYPI sources are allowed in requirements.txt"
        )

    result["packages"] = deps  # type: ignore[assignment]

    return result


def parse_req_value(
    file_content: str,
    extra_args: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    deps: Dict[str, str],
    np_deps: Dict[str, str],
    sys_deps: Dict[str, str],
) -> Optional[str]:
    python_version = None
    for line in file_content.splitlines():
        line = line.strip()
        if not line:
            continue
        splits = line.split(maxsplit=1)
        first_word = splits[0]
        if len(splits) > 1:
            rem = splits[1]
        else:
            rem = None
        if first_word in ("-i", "--index-url"):
            raise InvalidEnvironmentException(
                "To specify a base PYPI index, set `METAFLOW_CONDA_DEFAULT_PYPI_SOURCE; "
                "you can specify additional indices using --extra-index-url"
            )
        elif first_word == "--extra-index-url" and rem:
            sources.setdefault("pypi", []).append(rem)
        elif first_word in ("-f", "--find-links", "--trusted-host") and rem:
            extra_args.setdefault("pypi", []).append(" ".join([first_word, rem]))
        elif first_word in ("--pre", "--no-index"):
            extra_args.setdefault("pypi", []).append(first_word)
        elif first_word == "--conda-channel" and rem:
            sources.setdefault("conda", []).append(rem)
        elif first_word == "--conda-pkg":
            # Special extension to allow non-python conda package specification
            split_res = REQ_SPLIT_LINE.match(splits[1])
            if split_res is None:
                raise InvalidEnvironmentException(
                    "Could not parse conda package '%s'" % splits[1]
                )
            s = split_res.groups()
            if s[1] is None:
                np_deps[s[0].replace(" ", "")] = ""
            else:
                np_deps[s[0].replace(" ", "")] = s[1].replace(" ", "").lstrip("=")
        elif first_word == "--sys-pkg":
            # Special extension to allow the specification of system dependencies
            # (currently __cuda and __glibc)
            split_res = REQ_SPLIT_LINE.match(splits[1])
            if split_res is None:
                raise InvalidEnvironmentException(
                    "Could not parse system package '%s'" % splits[1]
                )
            s = split_res.groups()
            pkg_name = s[0].replace(" ", "")
            if pkg_name not in CONDA_SYS_DEPENDENCIES:
                raise InvalidEnvironmentException(
                    "System package '%s' not allowed. Values allowed are: %s"
                    % (pkg_name, str(CONDA_SYS_DEPENDENCIES))
                )
            if s[1] is None:
                raise InvalidEnvironmentException(
                    "System package '%s' requires a version" % pkg_name
                )
            sys_deps[pkg_name] = s[1].replace(" ", "").lstrip("=")
        elif first_word.startswith("#"):
            continue
        elif first_word.startswith("-"):
            raise InvalidEnvironmentException(
                "'%s' is not a supported line in a requirements.txt" % line
            )
        else:
            try:
                parsed_req = Requirement(line)
            except InvalidRequirement as ex:
                raise InvalidEnvironmentException("Could not parse '%s'" % line) from ex
            dep_name = parsed_req.name
            if parsed_req.extras:
                dep_name += "[%s]" % ",".join(parsed_req.extras)
            if parsed_req.url:
                dep_name += "@%s" % parsed_req.url
            specifier = str(parsed_req.specifier).lstrip(" =")
            if parsed_req.marker:
                specifier += ";" + str(parsed_req.marker)
            if dep_name == "python":
                if specifier:
                    python_version = specifier
            else:
                deps[dep_name] = specifier
    return python_version


def parse_yml_value(
    file_content: str,
    _: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    conda_deps: Dict[str, str],
    pypi_deps: Dict[str, str],
    sys_deps: Dict[str, str],
) -> Optional[str]:
    python_version = None  # type: Optional[str]
    mode = None
    for line in file_content.splitlines():
        if not line:
            continue
        elif line[0] not in (" ", "-"):
            line = line.strip()
            if line == "channels:":
                mode = "sources"
            elif line == "dependencies:":
                mode = "deps"
            elif line == "pypi-indices:":
                mode = "pypi_sources"
            else:
                mode = "ignore"
        elif mode and mode.endswith("sources"):
            line = line.lstrip(" -").rstrip()
            sources.setdefault("conda" if mode == "sources" else "pypi", []).append(
                line
            )
        elif mode and mode.endswith("deps"):
            line = line.lstrip(" -").rstrip()
            if not line:
                continue
            if line == "pip:":
                mode = "pypi_deps"
            elif line == "sys:":
                mode = "sys_deps"
            else:
                to_update = (
                    conda_deps
                    if mode == "deps"
                    else pypi_deps if mode == "pypi_deps" else sys_deps
                )
                splits = YML_SPLIT_LINE.split(line.replace(" ", ""), maxsplit=1)
                if len(splits) == 1:
                    if splits[0] != "python":
                        if mode == "sys_deps":
                            raise InvalidEnvironmentException(
                                "System package '%s' requires a version" % splits[0]
                            )

                        dep_name = splits[0]
                        if (
                            dep_name.startswith("/")
                            or dep_name.startswith("git+")
                            or dep_name.startswith("https://")
                            or dep_name.startswith("ssh://")
                        ):
                            # Handle the case where only the URL is specified
                            # without a package name
                            depname_and_maybe_tag = dep_name.split("/")[-1]
                            depname = depname_and_maybe_tag.split("@")[0]
                            if depname.endswith(".git"):
                                depname = depname[:-4]
                            dep_name = "%s@%s" % (depname, dep_name)
                        to_update[dep_name] = ""
                else:
                    dep_name, dep_operator, dep_version = splits
                    if dep_operator not in ("=", "=="):
                        if mode == "sys_deps":
                            raise InvalidEnvironmentException(
                                "System package '%s' requires a specific version not '%s'"
                                % (splits[0], dep_operator + dep_version)
                            )
                        dep_version = dep_operator + dep_version

                    if dep_name == "python":
                        if dep_version:
                            if python_version:
                                raise InvalidEnvironmentException(
                                    "Python versions specified multiple times in "
                                    "the YAML file."
                                )
                            python_version = dep_version
                    else:
                        if (
                            mode == "sys_deps"
                            and dep_name not in CONDA_SYS_DEPENDENCIES
                        ):
                            raise InvalidEnvironmentException(
                                "System package '%s' not allowed. Values allowed are: %s"
                                % (dep_name, str(CONDA_SYS_DEPENDENCIES))
                            )
                        to_update[dep_name] = dep_version

    return python_version


def parse_toml_value(
    file_content: str,
    _: Dict[str, List[str]],
    sources: Dict[str, List[str]],
    deps: Dict[str, str],
    __: Dict[str, str],
    ___: Dict[str, str],
) -> Optional[str]:
    try:
        import tomllib as toml  # Python 3.11+
    except ImportError:
        try:
            import tomli as toml  # Python < 3.11 (requires "tomli" package)
        except ImportError as e:
            raise InvalidEnvironmentException(
                "Could not import a TOML library. For Python <3.11, please install 'tomli'."
            ) from e

    data = toml.loads(file_content)

    project = data.get("project", {})
    requirements = project.get("dependencies", [])
    python_version = project.get("requires-python")
    for dep_line in requirements:
        try:
            parsed_req = Requirement(dep_line)
        except InvalidRequirement as ex:
            raise InvalidEnvironmentException("Could not parse '%s'" % dep_line) from ex

        dep_name = parsed_req.name
        if parsed_req.extras:
            dep_name += "[%s]" % ",".join(parsed_req.extras)
        if parsed_req.url:
            dep_name += "@%s" % parsed_req.url
        specifier = str(parsed_req.specifier).lstrip(" =")
        if parsed_req.marker:
            specifier += ";" + str(parsed_req.marker)

        if dep_name == "python":
            raise InvalidEnvironmentException(
                "Python specification should be specified as 'requires-python'"
            )
        deps[dep_name] = specifier

    # Also parse poetry sources as extra indices
    poetry_sources = data.get("tool", {}).get("poetry", {}).get("source", [])
    for s in poetry_sources:
        if "url" in s:
            sources.setdefault("pypi", []).append(s["url"])

    # Also parse uv index as extra indices
    uv_sources = data.get("tool", {}).get("uv", {}).get("index", [])
    for s in uv_sources:
        if "url" in s:
            sources.setdefault("pypi", []).append(s["url"])

    return python_version



import os
import tempfile

from pathlib import Path
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING

from metaflow import FlowMutator

from .conda_common_decorator import StepRequirementMixin
from .conda_flow_decorator import PackageRequirementFlowDecorator
from .parsers import req_parser, toml_parser
from .utils import (
    call_binary,
    determine_uv_env_python_version,
    compute_file_hash,
)


if TYPE_CHECKING:
    import metaflow


# ## resolved_uv, resolved_req, and resolved_env flow decorators.

# ### 1. TLDR:
# * resolved_uv is a convenient way to allow users to use their uv managed environment,
#   reflected by uv.lock and pyproject.toml files, in Metaflow flows.
# * resolved_req similarly, is a convenient way to allow users to use their *resolved*
#   requirements.txt
# * resolved_env is an both a base class for the previous one, and later to be an umbrella
#   decorator to automatically determine which scenario (uv.lock based or requirements.txt based)
#   the user is using.

# ### 2. Backgrounds

# #### 2.1 the uv movement
# This whole initiative started with uv, and evolved into a more general support for "resolved_env".

# uv is gaining traction as a fast dependency management tool.
# When developing, there are several major ways to use uv in the command.

# Our customers often use uv in the following ways:

# 1. "uv init" in a project folder to create a pyproject.toml file
# 2. "uv add dep" to add more dependencies
# 3. As they add, pyproject.toml is updated to contain the user specified
#     dependencies, and uv.lock is updated to contain the fully resolved
#     (derived) dependencies.
# 4. "uv sync" to sync (install and uninstall) the dependencies
# 5. "uv run" to run a python file with current project defined dependencies.

# And we decided to allow metaflow to directly consume uv's "state" (uv.lock)
# to create metaflow's flow running environment.

# #### 2.2. From uv to resolved_env
# As we progress on the uv initiative, we realized that a new pattern to support
# is emerging: "resolved environments". Contrary to unresolved environments where
# a resolution step is needed (e.g. conda/pip compiling a list of user dependencies
# into a locked expanded set of packages to install), resolved environments
# are already a locked set of packages to install.

# Another example is resolved requirements.txt files, which are often generated
# by from requirements.in, which is smaller set of user-defined dependencies.

# As a result, we decided to create resolved_env as a base class for both
# resolved_uv and resolved_req decorators, and later to be an umbrella decorator
# to automatically determine which scenario the user is using.

# ### 3. Usage (resolved_uv)

#     Users can simply do this:

#     ```
#     @resolved_uv
#     class MyFlow(FlowSpec):
#         ...
#     ```

#     And this uv decorator will locate uv.lock and pyproject.toml files
#     (searching from the flow code directory to its parent directories),
#     and create environment for a whole flow.


# ### 3. Usage (resolved_req)

#     Similarly, users can simply do this:

#     ```
#     @resolved_req
#     class MyFlow(FlowSpec):
#         ...
#     ```

#     And this decorator will
#     1. locate requirements.txt and requirements.in files, searching from the
#        flow code directory to its parent directories
#     2. confirm requirements.txt is resolved.
#     3. create environment for a whole flow

# ### 4. Caveats
#     1. resolved_uv and resolved_req decorators cannot be used together with
#        conda_base, pip_base, pypi_base, named_env_base, sys_packages_base
#        decorators on the same flow.
#     2. Steps in flows decorated by resolved_uv or resolved_req decorators
#        cannot use other environment decorators (for now -- this could be later
#        relaxed to support npconda decorator if we add something like this).
#     3. groups and optional dependencies in pyproject.toml are not supported yet.


class ResolvedEnvironmentBaseFlowMutator(FlowMutator):
    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)

        # The actual scenario computation (finding requirements.txt, uv.lock etc.) is done
        # at mutate() stage, not in init(), because mutable_flow object is not available
        # at init() stage. And we need mutable_flow to find the flow code path
        self.kwargs = kwargs

        # Get cwd at the init stage in case it later changes
        # at the mutate() stage, use_flow_dir parameter will be checked
        # and if the parameter is False, cwd rather than the flow code dir
        # will be used to start search for uv.lock/pyproject.toml etc.
        self.cwd = os.getcwd()

    def mutate(self, mutable_flow):
        if not self._do_mutate():
            return

        ResolvedEnvironmentBaseFlowMutator._verify_no_conflict_flow_decorators(
            mutable_flow
        )

    @staticmethod
    def _locate_first_folder_contains_any_file_from_list(
        file_name_no_path_list: List[str],
        search_start_path: str,
        # The first dir found containing any of the files will need
        # to contain all files if this flag is set to True.
        validate_all_files_in_same_dir: bool = True,
    ) -> Optional[str]:
        for p in [Path(search_start_path).absolute()] + list(
            Path(search_start_path).absolute().parents
        ):
            count = 0
            for file_name_no_path in file_name_no_path_list:
                if (p / file_name_no_path).exists():
                    count += 1
                    if not validate_all_files_in_same_dir:
                        return str(p)

            if count > 0:
                if validate_all_files_in_same_dir and count < len(
                    file_name_no_path_list
                ):
                    raise FileNotFoundError(
                        f"Found only partial files in the same directory {p}. "
                        + f"Expected files: {file_name_no_path_list}"
                    )
                return str(p)

        return None

    @staticmethod
    def _do_mutate() -> bool:
        # Returns true if this decorator should apply and false otherwise
        # It will return false when, for example, we already have the
        # environment setup and we don't really need to do anything else
        return os.environ.get("_METAFLOW_CONDA_ENV") is None

    @staticmethod
    def _mutate_step(
        step_name: str,
        step: "metaflow.user_decorators.mutable_step.MutableStep",
        pylock_toml_path: str,
        # See _parse_requirements_txt() for the format of this dict.
        user_deps: Dict[str, str],
        user_sources: List[str],
        # takes a full_id_unique_keys dict rather than uv_content_hash
        # to allow more flexibility in the future.
        full_id_unique_keys: Dict[str, str],
    ):
        ResolvedEnvironmentBaseFlowMutator._verify_no_conflict_step_decorators(
            step_name, step
        )
        step.add_decorator(
            "pylock_toml_internal",
            deco_kwargs={
                "path": pylock_toml_path,
                "user_deps_for_hash": user_deps,
                "user_sources_for_hash": user_sources,
                "full_id_unique_keys": full_id_unique_keys,
            },
        )

    @staticmethod
    def _verify_no_conflict_flow_decorators(mutable_flow):
        package_req_flow_deco_derived_classnames = (
            PackageRequirementFlowDecorator.get_derived_classes_fullnames()
        )

        conflicting_decorators = []
        for t in mutable_flow.decorator_specs:
            if t[1] in package_req_flow_deco_derived_classnames:
                conflicting_decorators.append(
                    package_req_flow_deco_derived_classnames[t[1]]
                )
        if conflicting_decorators:
            raise ValueError(
                "Conflicting flow decorators detected: "
                f"found {', '.join(conflicting_decorators)}. "
                "resolved_* decorators are not compatible with the following "
                "decorators: "
                f"{', '.join(package_req_flow_deco_derived_classnames.values())}. "
                "Please use only one of the resolved_* decorators per flow."
            )

    @staticmethod
    def _verify_no_conflict_step_decorators(
        step_name: str,
        step: "metaflow.user_decorators.mutable_step.MutableStep",
    ):

        req_mixin_derived_classes = StepRequirementMixin.get_derived_classes_fullnames()
        conflicting_decorators = []
        for t in step.decorator_specs:
            if t[1] in req_mixin_derived_classes:
                conflicting_decorators.append(req_mixin_derived_classes[t[1]])
        if conflicting_decorators:
            raise ValueError(
                f"Conflicting decorators detected on step {step_name}: "
                f"found {', '.join(conflicting_decorators)}. "
                "resolved_* decorators are not compatible with the following "
                "decorators: "
                f"{', '.join(req_mixin_derived_classes.values())}. "
                "Please use only one of the resolved_* decorators per flow."
            )


class ResolvedUVEnvFlowDecorator(ResolvedEnvironmentBaseFlowMutator):
    ENV_FILE_LIST = [
        "uv.lock",
        "pyproject.toml",
    ]
    PARAM_KEYS = [
        "use_flow_dir",
        "path",
    ]

    def init(self, *args, **kwargs):
        # Need to keep this method since the mutator needs to have its own one to be
        # called.
        super().init(*args, **kwargs)

    def mutate(self, mutable_flow):
        if not self._do_mutate():
            return

        super().mutate(mutable_flow)

        filtered_kwargs = {
            k: v
            for k, v in self.kwargs.items()
            if k in ResolvedUVEnvFlowDecorator.PARAM_KEYS and v is not None
        }
        if "path" in filtered_kwargs and "use_flow_dir" in filtered_kwargs:
            raise ValueError("Cannot set both `path` and `use_flow_dir` in resolved_uv")

        if "path" in filtered_kwargs:
            proj_config_dir = filtered_kwargs["path"]
        else:
            cwd = self.cwd
            if filtered_kwargs.get("use_flow_dir", True):
                import inspect

                cwd = os.path.dirname(inspect.getfile(mutable_flow._flow_cls))

            proj_config_dir = self._locate_first_folder_contains_any_file_from_list(
                self.ENV_FILE_LIST, cwd
            )

        # Ideally we should run this check at init() stage not at mutate() stage,
        # but mutable_flow is not available at init() stage.
        if not proj_config_dir:
            raise ValueError(
                "@resolved_uv used outside of a uv directory. "
                "Please specify a directory using the `path` argument "
                f"containing the files: {', '.join(self.ENV_FILE_LIST)}"
            )

        # We need to check if the files exist properly in case the user specifies
        # a specific path. In the other case, we have already located these files
        if any(
            not os.path.exists(os.path.join(proj_config_dir, filename))
            for filename in self.ENV_FILE_LIST
        ):
            raise FileNotFoundError(
                f"Cannot find all required files. Files {', '.join(self.ENV_FILE_LIST)} "
                f"must exist in the specified path {proj_config_dir}"
            )

        temp_toml_file_path, user_deps, user_sources, uv_lock_file_content_hash = (
            self._handle_uv_lock_scenario(
                os.path.join(proj_config_dir, "uv.lock"),
                os.path.join(proj_config_dir, "pyproject.toml"),
            )
        )

        for step_name, step in mutable_flow.steps:
            ResolvedEnvironmentBaseFlowMutator._mutate_step(
                step_name,
                step,
                temp_toml_file_path,
                user_deps,
                user_sources,
                full_id_unique_keys={
                    "uv_lock_content_hash": uv_lock_file_content_hash,
                },
            )

    # @staticmethod
    # def _handle_req_txt_scenario(
    #     requirements_txt_path: str, requirements_in_path: str
    # ) -> tuple[Optional[str], Dict[str, str]]:
    #     raise NotImplementedError("This will be supported in a stacked PR soon.")

    # Returns: (temp_pylock_toml_path, user_deps_dict, user_sources_list, uv_lock_content_hash)
    @staticmethod
    def _handle_uv_lock_scenario(
        uv_lock_path: str, pyproject_toml_path: str
    ) -> Tuple[str, Dict[str, str], List[str], str]:
        # 1. read pyproject.toml
        user_deps, user_srcs = (
            ResolvedUVEnvFlowDecorator._parse_pyproject_toml_to_user_deps_dict(
                pyproject_toml_path
            )
        )

        uv_env_python_version = determine_uv_env_python_version(
            str(Path(pyproject_toml_path).parent)
        )
        user_deps["python"] = uv_env_python_version

        # 2. call uv to generate a temp pylock.toml from uv.lock
        #  Note that this is just a translation, *not* a resolution.

        temp_dir = tempfile.mkdtemp()
        temp_path = str(Path(temp_dir) / "pylock.toml")
        cwd = os.path.dirname(uv_lock_path)

        # "uv export" command does not take an input uv.lock path,
        # it assumes uv.lock is in the current working directory.
        call_binary(
            ["export", "--frozen", "-o", temp_path],
            "uv",
            cwd=cwd,
        )

        # Calculate uv lock file content hash, to be passed to PylockToml Resolver,
        # and eventually ResolvedEnvironment, to be a component of full_id hash calculation,
        # and will be used to invalidate the cache if pyproject.toml is the same but uv.lock changes.
        uv_lock_content_hash = compute_file_hash(uv_lock_path)

        return temp_path, user_deps, user_srcs, uv_lock_content_hash

    @staticmethod
    def _parse_pyproject_toml_to_user_deps_dict(
        pyproject_toml_path: str,
    ) -> Tuple[Dict[str, str], List[str]]:
        with open(pyproject_toml_path, "r", encoding="utf-8") as fd:
            result = toml_parser(fd.read())

        # Eventually python will be one of the packages, and gets checked by
        # EnvsResolver._resolve_environments(), and built in to builder environment
        # packages. And python will be installed by conda, with the specified
        # python version.
        return result["packages"], result.get("extra_indices", [])


class ResolvedReqFlowDecorator(ResolvedEnvironmentBaseFlowMutator):
    """
    This is a placeholder for future support of resolved requirements.txt based flows.

    See ResolvedUVEnvFlowDecorator for more details about the rationale, usage,
    and under the hood design.
    """

    PARAM_KEYS = [
        "use_flow_dir",
        "path",
    ]
    ENV_FILE_LIST = [
        "requirements.txt",
        "requirements.in",
    ]

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        raise NotImplementedError("This will be supported soon.")

    def mutate(self, mutable_flow):
        if not self._do_mutate():
            return
        # Step 1: find requirements.txt and requirements.in
        # Step 2: confirm requirements.txt is resolved
        # Step 3: parse requirements.in to get user_deps dict
        # Step 4: call uv to translate requirements.txt to pylock.toml
        # Step 5: add pylock_toml_internal decorator to each step
        raise NotImplementedError("This will be supported soon.")

    # Sample return value:
    # {
    #     'pydub': '==0.25.1', 'numpy': '>=1.21.0', 'pandas': '>=1.3.0',
    #     "wrapt": "==1.17.3",
    #     "bdp-gandalf-policy-manager": "==~=0.1.9",
    #     "astroid": "==<=3.3.11",
    #     "blinker": "==<=2.0.0,>=1.9.0",
    # }
    #
    # Refer to test_parsing_requirements() in uv_decorator_test.py for more details.
    @staticmethod
    def _parse_requirements_txt(
        requirements_txt_path: str,
    ) -> Tuple[Dict[str, str], List[str]]:
        with open(requirements_txt_path, "r", encoding="utf-8") as fd:
            result = req_parser(fd.read())
        result["packages"]["python"] = result["python"]
        return result["packages"], result.get("extra_indices", [])



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import uuid
import time

from itertools import chain
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import (
    cast,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,  # noqa
    Tuple,
    Type,
    TYPE_CHECKING,
)

from metaflow.debug import debug
from metaflow.metaflow_config import (
    CONDA_DEPENDENCY_RESOLVER,
    CONDA_PREFERRED_FORMAT,
    CONDA_MIXED_DEPENDENCY_RESOLVER,
    CONDA_PYPI_DEPENDENCY_RESOLVER,
    CONDA_SYS_DEPENDENCIES,
    CONDA_SYS_DEFAULT_PACKAGES,
    CONDA_USE_REMOTE_LATEST,
)
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.system import _system_monitor, _system_logger

from metaflow._vendor.packaging.version import parse as parse_version
from metaflow.metaflow_config import PYLOCK_TOML_DEPENDENCY_RESOLVER
from .env_descr import (
    EnvID,
    EnvType,
    PackageSpecification,
    ResolvedEnvironment,
    env_type_for_deps,
)

from .resolvers import Resolver
from .utils import (
    CondaException,
    channel_or_url,
    arch_id,
    get_requirement,
    dict_to_strlist,
    filter_user_reqs_by_markers,
    get_builder_envs_dep,
    merge_dep_dicts,
    plural_marker,
    split_into_dict,
    tstr_to_dict,
)

if TYPE_CHECKING:
    from .conda import Conda


class EnvsResolver(object):
    def __init__(self, conda: "Conda"):
        # key: EnvID; value: dict containing:
        #  - "id": key
        #  - "steps": steps using this environment
        #  - "user_deps": array of requested dependencies
        #  - "deps": full array of dependencies -- typically user_deps except if base_env
        #  - "sources": full array of sources
        #  - "extras": additional arguments (typically used for PYPI)
        #  - "conda_format": list of formats we want the Conda packages in or ["_any"] if no pref
        #  - "base": optional base environment this environment derives from
        #  - "base_accurate": T/F; True if the "deps" for the base environment are known
        #  - "resolved": The resolved environment
        #  - "need_caching": T/F: True if we need to cache the environment
        #  - "force": T/F: True if we need to re-resolve
        # Log a unique UUID for each EnvsResolver instance
        self.uuid = str(uuid.uuid4())
        self._requested_envs = {}  # type: Dict[EnvID, Dict[str, Any]]
        self._builder_envs = {}  # type: Dict[EnvID, Dict[str, Any]]
        self._conda = conda
        self._non_step_envs = False
        self._co_resolved_force_resolve = set()  # type: Set[str]

    @staticmethod
    def find_resolved_environment(
        conda: "Conda",
        architecture: str,
        user_deps: Dict[str, List[str]],
        user_sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        base_env: Optional[ResolvedEnvironment] = None,
        force: bool = False,
        local_only: bool = False,
        use_latest: str = CONDA_USE_REMOTE_LATEST,
        full_id_unique_keys: Optional[Dict[str, str]] = None,
    ) -> Tuple[
        EnvType,
        EnvID,
        Dict[str, List[str]],
        Dict[str, List[str]],
        Dict[str, List[str]],
        Dict[str, List[str]],
        bool,
        Optional[ResolvedEnvironment],
    ]:
        """
        Determine the env_type, env_id, deps, compatible_base_env, and resolved_env,
        given the requested configuration.

        Returns
        -------
        Tuple[
            EnvType,        # env_type
            EnvID,          # env_id
            Dict[str, List[str]], # user_sources (possibly updated)
            Dict[str, List[str]], # user_deps (possibly updated)
            Dict[str, List[str]], # deps
            Dict[str, List[str]], # extras (possibly updated)
            bool,           # compatible_base_env
            Optional[ResolvedEnvironment], # resolved_env
        ]
        """
        if base_env:
            (
                env_type,
                env_id,
                user_sources_out,
                user_deps_out,
                deps,
                extras_out,
                compatible_base_env,
            ) = EnvsResolver.extract_info_from_base(
                conda, base_env, user_deps, user_sources, extras, architecture
            )
            if force:
                compatible_base_env = False
                debug.conda_exec(
                    "Force flag is set, ignoring compatible base environment"
                )
        else:
            compatible_base_env = False
            env_type = env_type_for_deps(user_deps)
            env_id = EnvID(
                ResolvedEnvironment.get_req_id(user_deps, user_sources, extras),
                "_default",
                architecture,
            )
            deps = user_deps
            user_sources_out = user_sources
            user_deps_out = user_deps
            extras_out = extras

        # Determine resolved_env
        resolved_env = None
        if (
            not force
            and base_env
            and base_env.env_id.req_id == env_id.req_id
            and base_env.env_id.arch == env_id.arch
        ):
            resolved_env = base_env
            # The env is not just compatible but it is the same. the compatible
            # flag is used to say: "this was almost the same env but not quite"
            # so we can determine if we cache or not
            compatible_base_env = False
        elif base_env and not force and compatible_base_env:
            debug.conda_exec(
                f"Base environment is compatible with the "
                "incoming dependencies, using it directly"
            )
            resolved_env = ResolvedEnvironment(
                user_dependencies=user_deps_out,
                user_sources=user_sources_out,
                user_extra_args=extras_out,
                arch=architecture,
                all_packages=list(base_env.packages),
                resolved_on=base_env.resolved_on,
                resolved_by=f"{base_env.resolved_by} (compatible with "
                f"{base_env.env_id.req_id}:{base_env.env_id.full_id})",
                co_resolved=base_env.co_resolved_archs,
                env_type=base_env.env_type,
                accurate_source=base_env.is_info_accurate,
                full_id_unique_keys=full_id_unique_keys,
            )
            resolved_env.dirty = True
        else:
            resolved_env = (
                conda.environment(env_id, local_only, use_latest, full_id_unique_keys)
                if not force
                else None
            )

        return (
            env_type,
            env_id,
            user_sources_out,
            user_deps_out,
            deps,
            extras_out,
            compatible_base_env,
            resolved_env,
        )

    def add_environment(
        self,
        architecture: str,
        user_deps: Dict[str, List[str]],
        user_sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        file_paths: Dict[str, List[str]] = {},
        step_name: str = "ad-hoc",
        base_env: Optional[ResolvedEnvironment] = None,
        base_from_full_id: bool = False,
        local_only: bool = False,
        use_latest: str = CONDA_USE_REMOTE_LATEST,
        force: bool = False,
        force_co_resolve: bool = False,
        full_id_unique_keys: Optional[Dict[str, str]] = None,
    ):
        """
        Add an environment to resolve to this EnvsResolver. The EnvsResolver will resolve
        the same environment only once and will mutualize anything it can.


        Parameters
        ----------
        architecture : str
            Architecture this environment is requested for
        user_deps : SDict[str, List[str]]
            The user dependencies of this environment -- if there is a base environment,
            these are the *additional* dependencies.
        user_sources : Dict[str, List[str]]
            The sources of this environment -- if there is a base environment, these are
            the *additional* sources.
        extras : Dict[str, List[str]]
            Any extra information to pass to Conda/Pypi (currently just pypi) -- if there
            is a base environment, these are the *additional* extras.
        step_name : str
            Step needing this environment or "ad-hoc" if not needed by a step
        base_env : Optional[ResolvedEnvironment], optional
            The base environment, if any, this environment is derived from, by default None
        base_from_full_id : bool, optional
            True if the base environment was extracted using just the full ID of the
            environment which makes it imprecise, by default False
        local_only : bool, optional
            True if we should only look for resolved environments locally and not
            remotely, by default False
        use_latest : str, default METAFLOW_CONDA_USE_REMOTE_LATEST
            For default environments, determine if we allow fetching the latest environment
            remotely resolved. If ":none:", we don't fetch anything. If ":username:", we
            fetch the latest environment resolved by the current user. If ":any:", we fetch
            the latest environment resolved by any user.
        force : bool, optional
            True if we should force resolution even if this environment is known,
            by default False
        force_co_resolve : bool, optional
            True if we should force co-resolution, by default False
        """
        self._non_step_envs = True

        # Check if we have the environment resolved already
        # TODO: We actually need to change this a little bit to effectively force `force`
        # to be True when the files listed in file_paths have changed. The logic can be
        # as follows:
        #  - we need to add an additional field to ResolvedEnvironment which would be
        #    addl_req_hash which would be a dictionary containing:
        #      - a "version" field
        #      - any other field to match -- for now, we can just make it a hash
        #  - the hash would be computed using the *hash* of the file types and the *hash*
        #    of the file contents
        #  - when doing environment, an environment returned would need to match the req_id
        #    (as usual) but also:
        #    - check that the addl_req_hash matches too.
        #  - We need to modify `environment` to accept this additional dictionary.
        #  - We would compute this dictionary here (TODO: check if there are other uses
        #    of it that would require us to compute the dictionary as well) and compare
        #    it with the addl_req_hash of the environment (make sure it's the same version
        #    and the same value). I would add the computation itself directly in the
        #    ResolvedEnvironment class so we can add versions of it later. Something like:
        #    def compute_addl_req_hash(data: Any):
        #      # Do stuff
        #      return {"version": 1, "value": "some_hash"}
        #    And similarly:
        #    def compare_addl_req_hash_version(self, data: Dict[str, Any]):
        #      if data["version"] != self.addl_req_hash_version # Or whatever
        #        return False
        #      return getattr(self, "compare_addl_req_hash_v{data['version']}")(data)
        #
        #    -
        (
            env_type,
            env_id,
            user_sources,
            user_deps,
            deps,
            extras,
            compatible_base_env,
            resolved_env,
        ) = self.find_resolved_environment(
            self._conda,
            architecture,
            user_deps,
            user_sources,
            extras,
            base_env=base_env,
            force=force,
            local_only=local_only,
            use_latest=use_latest,
            full_id_unique_keys=full_id_unique_keys,
        )

        # Check if we have already requested this environment
        if env_id not in self._requested_envs:
            if force_co_resolve:
                if resolved_env is None:
                    # Invalidate any previously resolved environment with the same req_id
                    for other_env_id, other_env in self._requested_envs.items():
                        if other_env_id.req_id == env_id.req_id:
                            other_env["resolved"] = None
                            other_env["need_caching"] = True
                    self._co_resolved_force_resolve.add(env_id.req_id)
                elif env_id.req_id in self._co_resolved_force_resolve:
                    # If another environment for the same req-id is not resolved, we
                    # need to resolve this one too.
                    resolved_env = None

            self._requested_envs[env_id] = {
                "id": env_id,
                "steps": [step_name],
                "user_deps": user_deps,
                "deps": deps,
                "sources": user_sources,
                "extras": extras,
                "file_paths": file_paths,
                "full_id_unique_keys": full_id_unique_keys,
                "conda_format": (
                    [CONDA_PREFERRED_FORMAT]
                    if CONDA_PREFERRED_FORMAT and CONDA_PREFERRED_FORMAT != "none"
                    else ["_any"]
                ),
                "base": base_env,
                "base_accurate": base_env
                and base_env.is_info_accurate
                and not base_from_full_id,
                "resolved": None if force else resolved_env,
                "need_caching": (
                    compatible_base_env if resolved_env is not None else True
                ),
                "env_type": env_type,
                "force": force,
            }

            debug.conda_exec(
                "Added environment to resolve %s" % str(self._requested_envs[env_id])
            )
        else:
            self._requested_envs[env_id]["steps"].append(step_name)
            debug.conda_exec(
                "Environment '%s' is also needed by '%s'" % (env_id, step_name)
            )

    def resolve_environments(self, echo: Callable[..., None]):
        """
        Resolve the environments added using `add_environment` and `add_environment_for_step`.

        Parameters
        ----------
        echo : Callable[..., None]
            Method to use to print things to the console
        """
        with _system_monitor.measure("metaflow.conda.all_resolve"):
            start = time.time()
            # At this point, we check in our backend storage if we have the files we need
            need_resolution = [
                env_id
                for env_id, req in self._requested_envs.items()
                if req["resolved"] is None
            ]
            if debug.conda:
                debug.conda_exec("Resolving environments:")
                for env_id in need_resolution:
                    info = self._requested_envs[env_id]
                    debug.conda_exec(
                        "%s (%s): %s" % (env_id.req_id, env_id.full_id, str(info))
                    )
            if len(need_resolution):
                self._resolve_environments(echo, need_resolution)
            # NOTE: Co-resolved environments allow you to resolve a bunch of "equivalent"
            # environments for different platforms. This is great as it can allow you to
            # run code on Linux and then instantiate an environment to look at it on Mac.
            # One issue though is that the set of packages on Linux may change while those
            # on mac may not (or vice versa) so it is possible to get in the following
            # situation:
            # - Co-resolve at time A:
            #   - Get linux full_id 123 and mac full_id 456
            # - Co-resolve later at time B:
            #   - Get linux full_id 123 and mac full_id 789
            # This is a problem because now the 1:1 correspondence between co-resolved
            # environments (important for figuring out which environment to use) is broken
            #
            # To solve this problem, we consider that co-resolved environments participate
            # in the computation of the full_id (basically a concatenation of all packages
            # across all co-resolved environments). This maintains the 1:1 correspondence.
            # It has a side benefit that we can use that same full_id for all co-resolved
            # environment making one easier to find from the other (instead of using indirect
            # links)
            co_resolved_envs = (
                {}
            )  # type: Dict[str, List[Tuple[EnvID, ResolvedEnvironment]]]
            for env_id, resolved_env, _, _ in self.need_caching_environments():
                co_resolved_envs.setdefault(env_id.req_id, []).append(
                    (env_id, resolved_env)
                )
            # Now we know all the co-resolved environments so we can compute the full
            # ID for all those environments
            for envs in co_resolved_envs.values():
                if len(envs) > 1:
                    ResolvedEnvironment.set_coresolved_full_id([x[1] for x in envs])
            if not self._non_step_envs:
                # Print some breadcrumbs about environments used in the flow.
                # This helps data-scientists understand a bit better what
                # environment is being used.
                newly_resolved_steps = sorted(
                    chain(
                        req["steps"]
                        for req in self._requested_envs.values()
                        if req["need_caching"]
                    )
                )
                previously_resolved_envs = {
                    sorted(req["steps"]): req["resolved"]
                    for req in self._requested_envs.values()
                    if not req["need_caching"]
                }
                if len(newly_resolved_steps):
                    end_line = "a newly resolved environment"
                else:
                    end_line = "newly resolved environments"

                echo(
                    f"    Step{plural_marker(len(newly_resolved_steps))} "
                    f"{', '.join(newly_resolved_steps)} using {end_line}."
                )
                for steps, resolved in previously_resolved_envs.items():
                    resolved = cast(ResolvedEnvironment, resolved)
                    echo(
                        f"    Step{plural_marker(len(steps))} "
                        f"{', '.join(steps)} using an environment previously "
                        f"resolved by {resolved.resolved_by} on {resolved.resolved_on} "
                        f"(id {resolved.env_id.full_id})."
                    )

            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="all_envs_resolved",
                payload={
                    "msg": str(
                        {
                            "resolve_time": time.time() - start,
                            "resolver_uuid": self.uuid,
                        }
                    ),
                    # Override the log stream to be the default metrics stream
                    "log_stream": "metrics",
                },
            )

    def all_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, Optional[ResolvedEnvironment], List[str]]]:
        """
        Returns an iterator over all the environments (resolved or not) this
        EnvsResolver knows about

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PYPI environments,
            we sometimes need to build another environment to build PYPI packages
            and/or resolve the PYPI dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, Optional[ResolvedEnvironment], List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment (if it was resolved)
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                yield env_id, cast(
                    Optional[ResolvedEnvironment], req["resolved"]
                ), cast(List[str], req["steps"])

    def resolved_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]:
        """
        Same as `all_environments` except for resolved environments only

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PYPI environments,
            we sometimes need to build another environment to build PYPI packages
            and/or resolve the PYPI dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                if req["resolved"] is not None:
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["steps"]
                    )

    def non_resolved_environments(
        self,
    ) -> Iterator[Tuple[EnvID, List[str]]]:
        """
        The opposite of `resolved_environments`. It never includes builder environments
        as those are always resolved

        Yields
        ------
        Iterator[Tuple[EnvID, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the list of step names this environment pertains to.
        """
        for env_id, req in self._requested_envs.items():
            if req["resolved"] is None:
                yield env_id, cast(List[str], req["steps"])

    def need_caching_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str], List[str]]]:
        """
        Same as `resolved_environments` but further limits the output to environments
        that need to be cached.

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PYPI environments,
            we sometimes need to build another environment to build PYPI packages
            and/or resolve the PYPI dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str], List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the Conda formats that need caching
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                # Resolved environments that were not already resolved or not fully
                # cached
                if req["resolved"] is not None and (
                    req["need_caching"]
                    or not req["resolved"].is_cached({"conda": req["conda_format"]})
                ):
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["conda_format"]
                    ), cast(List[str], req["steps"])

    def new_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]:
        """
        Same as `resolved_environments` but further limits the output to environments
        that were newly resolved (ie: excludes ones that were already resolved prior
        to starting this EnvsResolver)

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PYPI environments,
            we sometimes need to build another environment to build PYPI packages
            and/or resolve the PYPI dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                # Resolved environments that were not already resolved
                if req["resolved"] is not None and req["need_caching"]:
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["steps"]
                    )

    def _resolve_environments(
        self, echo: Callable[..., None], env_ids: Sequence[EnvID]
    ):
        start = time.time()
        if len(env_ids) == len(self._requested_envs):
            echo(
                "    Resolving %d environment%s %s..."
                % (
                    len(env_ids),
                    plural_marker(len(env_ids)),
                    "in flow " if not self._non_step_envs else "",
                ),
                nl=False,
            )
        else:
            echo(
                "    Resolving %d of %d environment%s %s(others are cached) ..."
                % (
                    len(env_ids),
                    len(self._requested_envs),
                    plural_marker(len(self._requested_envs)),
                    "in flow " if not self._non_step_envs else "",
                ),
                nl=False,
            )

        # In PYPI-ONLY scenarios, we actually need a sub-environment to properly resolve
        # the pypi environment (due to https://github.com/pypa/pip/issues/11664).
        # We try to mutualize these environments.
        # There are other cases where we need a builder env but we only build it
        # for user in PYPI-ONLY scenarios because we are not sure we need it in other
        # cases. If we do need it in the other cases, it will be lazily built and returned
        # so we can cache them for later use (they just won't be re-used and cached
        # this time so it is possible that we resolve the same builder environment
        # multiple times). This happens, for example, in mixed mode when we need to build
        # PYPI packages from source.

        # In the case of cross architecture build we need two things:
        #  - a "vanilla" builder environment for this machine's architecture
        #    (we do not need the npconda packages in
        #    that since we won't use them for anything) -- this will be the actual
        #    "builder" environment
        #  - a base conda environment for the target architecture (which includes the
        #    npconda packages). We will use this as the base environment.
        # For simplicity, we call both of them the "builder" environments (the first is
        # an environment with which we build the pypi environment and the second is the
        # building block of the final environment).

        # Figure out the set of builder environments we need to build. The key is the
        # requirement ID for which we are building these builder environments and the
        # value is the envid of the environments built for it.
        builders_by_req_id = {}  # type: Dict[str, List[EnvID]]
        my_arch = arch_id()
        env_ids_to_resolve = []
        for env_id in env_ids:
            env_type = self._requested_envs[env_id].get("env_type")
            env_ids_to_resolve.append(env_id)
            if env_type is None:
                env_type = env_type_for_deps(self._requested_envs[env_id]["deps"])

            to_resolve_info = self._requested_envs[env_id]

            # Extract just the conda sources
            builder_sources = {"conda": to_resolve_info["sources"].get("conda", [])}

            # Figure out the python dependency
            python_dep = [
                d
                for d in to_resolve_info["user_deps"].get("conda", [])
                if d.startswith("python==")
            ]
            python_version_requested = python_dep[0].split("==")[1]
            # A common pattern is people request a Python version using "major.minor"
            # instead of "major.minor.*". Which is likely what they intend (ie: any
            # version of that major.minor) so we convert it to the more common format.
            if python_version_requested.count(".") == 1:
                python_version_requested = python_version_requested + ".*"
                python_dep = [f"python=={python_version_requested}"]

            if python_version_requested[0] not in ("~", "<", ">", "!", "="):
                python_version_requested = "python==%s" % python_version_requested
            else:
                python_version_requested = "python%s" % python_version_requested

            # We do this lazily here
            self._requested_envs[env_id][
                "python_version_requested"
            ] = python_version_requested

            if not EnvsResolver._env_needs_builder_env(self._requested_envs[env_id]):
                continue

            # Make sure we have everything we need in the builder environment
            python_dep.extend(
                get_builder_envs_dep(
                    [
                        d
                        for d in to_resolve_info["user_deps"].get("conda", [])
                        if not d.startswith("python==")
                    ]
                )
            )

            # The user dependencies are everything in npconda, sys and the pythondep
            builder_user_deps = {
                env_id.arch: {
                    "conda": python_dep,
                    "npconda": to_resolve_info["user_deps"].get("npconda", []),
                    "sys": to_resolve_info["user_deps"].get("sys", []),
                }
            }
            if env_id.arch != my_arch:
                # Here we need an extra "vanilla" environment with just the python
                # dependency
                builder_user_deps[my_arch] = {"conda": python_dep}

            for builder_arch, builder_deps in builder_user_deps.items():
                builder_env_id = EnvID(
                    ResolvedEnvironment.get_req_id(builder_deps, builder_sources, {}),
                    "_default",
                    builder_arch,
                )
                if builder_env_id in self._builder_envs:
                    builder_env_info = self._builder_envs[builder_env_id]
                    builder_env_info["steps"].extend(to_resolve_info["steps"])
                else:
                    builder_env = None
                    if not to_resolve_info["force"]:
                        builder_env = self._conda.environment(builder_env_id)
                    builder_env_info = {
                        "id": builder_env_id,
                        "steps": to_resolve_info["steps"],
                        "user_deps": builder_deps,
                        "deps": builder_deps,
                        "python_version_requested": python_version_requested,
                        "sources": builder_sources,
                        "file_paths": {},
                        "extras": {},
                        "conda_format": to_resolve_info["conda_format"],
                        "base": None,
                        "base_accurate": False,
                        "resolved": builder_env,
                        "need_caching": builder_env is None,
                        "env_type": EnvType.CONDA_ONLY,
                        "full_id_unique_keys": {},
                    }
                self._builder_envs[builder_env_id] = builder_env_info
                builders_by_req_id.setdefault(env_id.req_id, []).append(builder_env_id)

        # Resolve all builder environments first so we can share them across steps
        if len(self._builder_envs):
            debug.conda_exec("Builder environments: %s" % str(self._builder_envs))

            # Build the environments
            with ThreadPoolExecutor() as executor:
                resolution_result = [
                    executor.submit(self._resolve, v, None)
                    for v in self._builder_envs.values()
                    if v["resolved"] is None
                ]
                for f in as_completed(resolution_result):
                    env_id, resolved, _ = f.result()
                    self._builder_envs[env_id]["resolved"] = resolved

        # Now that we have all the builder environments, we go ahead and resolve the
        # actual environments. We can pass down the builder environments as needed

        if len(env_ids_to_resolve):
            with ThreadPoolExecutor() as executor:
                resolution_result = [
                    executor.submit(self._resolve, v, builders_by_req_id)
                    for k, v in self._requested_envs.items()
                    if k in env_ids_to_resolve
                ]
                for f in as_completed(resolution_result):
                    env_id, resolved_env, addl_builder_envs = f.result()

                    resolved_env_id = resolved_env.env_id
                    cached_resolved_env = self._conda.environment(resolved_env_id)
                    # This checks if there is the same resolved environment already
                    # cached (in which case, we don't have to check a bunch of things
                    # so makes it nicer)
                    if cached_resolved_env:
                        resolved_env = cached_resolved_env

                    self._requested_envs[env_id]["resolved"] = resolved_env
                    debug.conda_exec(
                        "For environment %s (%s) %s need packages %s"
                        % (
                            resolved_env_id.req_id,
                            resolved_env_id.full_id,
                            self._requested_envs[env_id],
                            ", ".join(
                                sorted([p.filename for p in resolved_env.packages])
                            ),
                        )
                    )
                    # Record any other builder environment that was built so we can
                    # properly cache it later
                    if addl_builder_envs is None:
                        addl_builder_envs = []
                    addl_builder_envs = [
                        env
                        for env in addl_builder_envs
                        if env.env_id not in self._builder_envs
                    ]
                    for addl_builder_env in addl_builder_envs:
                        # We "hack" the extract_from_base to get the proper user and
                        # full requirements from addl_builder_env
                        _, _, _, user_deps, full_deps, _, _ = (
                            self.extract_info_from_base(
                                self._conda,
                                addl_builder_env,
                                {},
                                {},
                                {},
                                addl_builder_env.env_id.arch,
                            )
                        )
                        self._builder_envs[addl_builder_env.env_id] = {
                            "id": addl_builder_env.env_id,
                            "steps": self._requested_envs[env_id]["steps"],
                            "user_deps": user_deps,
                            "deps": full_deps,
                            "sources": addl_builder_env.sources,
                            "extras": {},
                            "conda_format": self._requested_envs[env_id][
                                "conda_format"
                            ],
                            "base": None,
                            "base_accurate": False,
                            "resolved": addl_builder_env,
                            "need_caching": True,
                            "env_type": EnvType.CONDA_ONLY,
                        }

        duration = int(time.time() - start)
        echo(" done in %d second%s." % (duration, plural_marker(duration)))

    def _resolve(
        self,
        env_desc: Mapping[str, Any],
        builder_environments: Optional[Dict[str, List[EnvID]]],
    ) -> Tuple[EnvID, ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        with _system_monitor.measure("metaflow.conda.resolve"):
            start = time.time()
            env_id = cast(EnvID, env_desc["id"])
            if builder_environments is None:
                builder_environments = {}

            builder_envs: Optional[List[ResolvedEnvironment]] = [
                self._builder_envs[builder_env_id]["resolved"]
                for builder_env_id in builder_environments.get(env_id.req_id, [])
            ]

            # Figure out the env_type
            env_type = cast(
                EnvType, env_desc.get("env_type") or env_type_for_deps(env_desc["deps"])
            )

            # Create the resolver object
            resolver = self.get_resolver_cls(env_desc)(self._conda)

            python_version_requested = env_desc["python_version_requested"]
            # Resolve the environment
            if env_type == EnvType.PYPI_ONLY:
                # Pypi only mode
                # In this mode, we also allow (as a workaround for poor support for
                # more advanced options in conda-lock (like git repo, local support,
                # etc)) the inclusion of conda packages that are *not* python packages.
                # To ensure this, we check the npconda packages, create an environment
                # for it and check if that environment doesn't contain python deps.
                # If that is the case, we then create the actual environment including
                # both conda and npconda packages and re-resolve. We could maybe
                # optimize to not resolve from scratch twice but given this is a rare
                # situation and the cost is only during resolution, it doesn't seem
                # worth it.
                npconda_deps = env_desc["deps"].get("npconda", [])
                if npconda_deps:
                    npcondaenv, _ = Resolver.get_resolver(CONDA_DEPENDENCY_RESOLVER)(
                        self._conda
                    ).resolve(
                        EnvType.CONDA_ONLY,
                        python_version_requested,
                        {"npconda": npconda_deps},
                        env_desc["sources"],
                        {},
                        env_id.arch,
                    )
                    if any(
                        (p.filename.startswith("python-") for p in npcondaenv.packages)
                    ):
                        raise InvalidEnvironmentException(
                            "Cannot specify a non-python Conda dependency that uses "
                            "python: %s. Please use the mixed mode instead."
                            % ", ".join([d for d in npconda_deps])
                        )
            resolved_env, builder_envs = resolver.resolve(
                env_type,
                python_version_requested,
                env_desc["deps"],
                env_desc["sources"],
                env_desc["extras"],
                env_id.arch,
                builder_envs,
                env_desc["base"],
                env_desc["file_paths"],
                env_desc["full_id_unique_keys"],
            )

            if env_desc["base"]:
                # We try to copy things over from the base environment as it contains
                # potential caching information we don't need to rebuild. This also
                # properly sets the user dependencies to what we want as opposed to
                # including everything we resolved for.
                merged_packages = []  # type: List[PackageSpecification]
                base_packages = {
                    p.filename: p.to_dict()
                    for p in cast(ResolvedEnvironment, env_desc["base"]).packages
                }
                for p in resolved_env.packages:
                    existing_info = base_packages.get(p.filename)
                    if existing_info:
                        merged_packages.append(
                            PackageSpecification.from_dict(existing_info)
                        )
                    else:
                        merged_packages.append(p)
                resolved_env = ResolvedEnvironment(
                    env_desc["user_deps"],
                    env_desc["sources"],
                    env_desc["extras"],
                    env_id.arch,
                    all_packages=merged_packages,
                    env_type=resolved_env.env_type,
                    accurate_source=env_desc["base_accurate"],
                )
            _system_logger.log_event(
                level="info",
                module="nflx.conda",
                name="env_resolved",
                payload={
                    "qualifier_name": str(env_id),
                    "msg": "Environment resolved in %d seconds" % (time.time() - start),
                    # Override the log stream to be the default metrics stream
                    "log_stream": "metrics",
                },
            )
            return env_id, resolved_env, builder_envs

    @staticmethod
    def extract_info_from_base(
        conda: "Conda",
        base_env: ResolvedEnvironment,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
    ) -> Tuple[
        EnvType,
        EnvID,
        Dict[str, List[str]],
        Dict[str, List[str]],
        Dict[str, List[str]],
        Dict[str, List[str]],
        bool,
    ]:
        """
        Returns the information about an environment extended from a given environment
        (in other words, taking the base environment first and then extending it
        with other dependencies).

        We try to be a bit smart about it -- if we realize that the base environment
        is compatible with the incoming dependencies, we will indicate this fact so
        that we can possibly avoid re-resolving the environment (which can be
        expensive, slow and lead to some weird bugs at times).

        Parameters
        ----------
        base_env : ResolvedEnvironment
            The base environment to use
        deps : Dict[str, List[str]]
            The user dependencies to add/update
        sources : Dict[str, List[str]]
            The sources to add/update
        extras : Dict[str, List[str]]
            Additional information passed to Conda/Pypi
        architecture : str
            The architecture to resolve for

        Returns
        -------
        Tuple[EnvType, EnvID, Dict[str, List[str]], Dict[str, List[str]],Dict[str, List[str]], Dict[str, List[str]], bool]
            The tuple contains:
                - the EnvType
                - the new EnvID
                - the set of sources
                - the set of user dependencies
                - the exact set of dependencies
                - the set of extras
                - True if the base environment is compatible with the incoming dependencies
        """
        if architecture != base_env.env_id.arch:
            raise CondaException(
                "Mismatched architecture when extending an environment"
            )
        # We form the new list of dependencies based on the ones we have in cur_env
        # and the new ones.
        # We need to split things up and reform
        # incoming_* is basically stuff that comes from everything outside of the
        # base_env
        python_package = next(
            (p for p in base_env.packages if p.package_name == "python"), None
        )
        python_version = python_package.package_version if python_package else None
        if python_version is None:
            raise CondaException("Python package not found in base environment")
        filtered_deps = filter_user_reqs_by_markers(deps, python_version, architecture)

        incoming_conda_deps = split_into_dict(filtered_deps.get("conda", []))
        incoming_npconda_deps = split_into_dict(filtered_deps.get("npconda", []))
        incoming_pypi_deps = split_into_dict(filtered_deps.get("pypi", []))
        incoming_sys_deps = split_into_dict(filtered_deps.get("sys", []))

        base_deps = tstr_to_dict(base_env.deps)
        base_sources = tstr_to_dict(base_env.sources)
        base_extras = tstr_to_dict(base_env.extras)

        user_conda_deps = merge_dep_dicts(
            split_into_dict(base_deps.get("conda", [])),
            incoming_conda_deps,
        )
        user_pypi_deps = merge_dep_dicts(
            split_into_dict(base_deps.get("pypi", [])),
            incoming_pypi_deps,
        )
        user_npconda_deps = merge_dep_dicts(
            split_into_dict(base_deps.get("npconda", [])),
            incoming_npconda_deps,
        )

        # Special handling of sys deps -- it specifies things like __cuda or __glibc.
        # In both cases, it specifies the maximum allowed version so we take the
        # smallest.

        base_sys_deps = split_into_dict(base_deps.get("sys", []))
        unknown_base_deps = set(base_sys_deps.keys()).difference(CONDA_SYS_DEPENDENCIES)
        if unknown_base_deps:
            raise CondaException(
                "Unhandled sys deps: %s" % ", ".join(unknown_base_deps)
            )
        unknown_incoming_deps = set(incoming_sys_deps.keys()).difference(
            CONDA_SYS_DEPENDENCIES
        )
        if unknown_incoming_deps:
            raise CondaException(
                "Unhandled sys deps: %s" % ", ".join(unknown_incoming_deps)
            )

        # We modify this later so make sure to copy it.
        user_sys_deps = dict(
            cast(Dict[str, str], CONDA_SYS_DEFAULT_PACKAGES.get(architecture, {}))
        )
        debug.conda_exec(
            "Default system deps are: %s; base env: %s; incoming: %s"
            % (user_sys_deps, base_sys_deps, incoming_sys_deps)
        )
        if base_sys_deps or incoming_sys_deps:
            for dep in CONDA_SYS_DEPENDENCIES:
                dep_str: str = cast(str, dep)
                v1 = base_sys_deps.get(dep_str, "9999")
                v2 = incoming_sys_deps.get(dep_str, "9999")
                if "=" in v1:
                    v1, v1_build = v1.split("=", 1)
                else:
                    v1_build = None
                if "=" in v2:
                    v2, v2_build = v2.split("=", 1)
                else:
                    v2_build = None
                v_min, v_min_build = v1, v1_build
                if parse_version(v2) < parse_version(v1):
                    v_min, v_min_build = v2, v2_build
                if v_min != "9999":
                    user_sys_deps[dep_str] = (
                        "%s=%s" % (v_min, v_min_build) if v_min_build else v_min
                    )

        for category, extra in base_extras.items():
            extras.setdefault(category, []).extend(extra)

        conda_deps = {
            p.package_name_with_channel(): p.package_version
            for p in base_env.packages
            if p.TYPE == "conda"
        }

        pypi_deps = {
            p.package_name: p.package_version
            for p in base_env.packages
            if p.TYPE == "pypi"
        }
        # Check if the incoming dependencies are already included and
        # compatible with the base environment. If so, we may be able to avoid
        # re-resolving the environment.
        compatible_base_env = True
        debug.conda_exec(
            "Base sys deps: %s; User sys deps: %s" % (base_sys_deps, user_sys_deps)
        )
        if base_sys_deps != user_sys_deps:
            compatible_base_env = False
        else:
            for dep_type, incoming_deps_per_type in filtered_deps.items():
                if dep_type in ("conda", "npconda"):
                    dep_dict = conda_deps
                elif dep_type == "pypi":
                    dep_dict = pypi_deps
                else:
                    continue  # These are the system deps

                for dep_req in incoming_deps_per_type:
                    req = get_requirement(dep_req)
                    if req.name not in dep_dict or (
                        len(req.specifier) != 0  # Make sure we have a version
                        and not req.specifier.contains(dep_dict[req.name])
                    ):
                        if len(req.specifier) == 0:
                            debug.conda_exec(
                                "Dependency %s prevents compatible base (not found)"
                                % req.name
                            )
                        else:
                            debug.conda_exec(
                                "Dependency %s prevents compatible base "
                                "(%s incompatible with requested version %s)"
                                % (
                                    req.name,
                                    dep_dict.get(req.name, "<not present>"),
                                    str(req.specifier),
                                )
                            )
                        compatible_base_env = False
                        break
                if not compatible_base_env:
                    break

        if not compatible_base_env:
            # If envs are compatible, the deps are the same so we don't change.
            conda_deps = merge_dep_dicts(
                merge_dep_dicts(conda_deps, incoming_conda_deps), incoming_npconda_deps
            )
            pypi_deps = merge_dep_dicts(pypi_deps, incoming_pypi_deps)

        # The dependencies we need to resolve for are:
        #  - the packages from the base environment
        #  - the packages the user requested
        deps = {
            "conda": dict_to_strlist(conda_deps),
            "npconda": dict_to_strlist(user_npconda_deps),
            "pypi": dict_to_strlist(pypi_deps),
            "sys": dict_to_strlist(user_sys_deps),
        }

        # The user requested dependencies are the ones that were requested for the
        # base environment as well as the ones that were added here
        user_deps = {
            "conda": dict_to_strlist(user_conda_deps),
            "npconda": list(deps["npconda"]),
            "pypi": dict_to_strlist(user_pypi_deps),
            "sys": list(deps["sys"]),
        }

        if base_env.env_type == EnvType.PYPI_ONLY and len(incoming_conda_deps) == 0:
            env_type = EnvType.PYPI_ONLY
        else:
            env_type = env_type_for_deps(deps)

        # Sources will be:
        #   - for Conda: base_sources, then user sources then default sources
        #   - for Pypi: default sources then base_sources then user sources

        new_sources = {
            "conda": list(
                dict.fromkeys(
                    map(
                        channel_or_url,
                        chain(
                            base_sources.get("conda", []),
                            sources.get("conda", []),
                            conda.default_conda_channels,
                        ),
                    )
                )
            )
        }

        if env_type != EnvType.CONDA_ONLY:
            new_sources["pypi"] = list(
                dict.fromkeys(
                    chain(
                        conda.default_pypi_sources,
                        base_sources.get("pypi", []),
                        sources.get("pypi", []),
                    )
                )
            )

        new_env_id = EnvID(
            ResolvedEnvironment.get_req_id(user_deps, new_sources, extras),
            "_default",
            architecture,
        )

        if (
            new_env_id.req_id == base_env.env_id.req_id
            and new_env_id.arch == base_env.env_id.arch
        ):
            # This means that the environment is identical to the base one. We can
            # set the full_id to be the same
            new_env_id = EnvID(
                new_env_id.req_id, base_env.env_id.full_id, new_env_id.arch
            )

        return (
            env_type,
            new_env_id,
            new_sources,
            user_deps,
            deps,
            extras,
            compatible_base_env,
        )

    @staticmethod
    def get_resolver_cls(env_desc: Mapping[str, Any]) -> Type[Resolver]:
        env_type = env_desc.get("env_type") or env_type_for_deps(env_desc["deps"])

        if env_desc.get("file_paths", {}).get("pylock_toml"):
            resolver_name = PYLOCK_TOML_DEPENDENCY_RESOLVER
        elif env_type == EnvType.CONDA_ONLY:
            resolver_name = CONDA_DEPENDENCY_RESOLVER
        elif env_type == EnvType.PYPI_ONLY:
            resolver_name = CONDA_PYPI_DEPENDENCY_RESOLVER
        elif env_type == EnvType.MIXED:
            resolver_name = CONDA_MIXED_DEPENDENCY_RESOLVER
        else:
            raise CondaException("Unhandled environment type %s" % env_type.value)

        if resolver_name is None or resolver_name == "none":
            raise CondaException(
                "Cannot resolve environments in %s mode because no resolver is configured"
                % env_type.value
            )
        return Resolver.get_resolver(resolver_name)

    @staticmethod
    def _env_needs_builder_env(env_desc: Mapping[str, Any]) -> bool:
        # Notes: this methods calculates the resolver class needed to look up REQUIRES_BUILDER_ENV,
        # and this calculation will happen again inside _resolve(). This inefficiency is ignored as
        # it a fast computation.
        return EnvsResolver.get_resolver_cls(env_desc).REQUIRES_BUILDER_ENV
